﻿using System;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Threading.Tasks;
using Inedo.Agents.Service;

namespace Inedo.Agents
{
    internal sealed class ClientConnection : AgentConnection
    {
        private MessageFormatter messageFormatter = new MessageFormatter.StandardMessageFormatter();

        public ClientConnection(Stream stream)
            : base(stream)
        {
        }

        public EndPoint TcpEndPoint { get; set; }

        public async Task<int> ReceiveHandshake()
        {
            var buffer = new byte[16];
            int read = await this.Stream.ReadBlockAsync(buffer, 0, buffer.Length);
            if (read < buffer.Length)
                throw new AgentConnectionException("Invalid handshake received.");

            if (ProtocolGuid != new Guid(buffer))
                throw new AgentConnectionException("Invalid handshake received.");

            read = await this.Stream.ReadBlockAsync(buffer, 0, 8);
            if (read < 8)
                throw new AgentConnectionException("Invalid handshake received.");

            int protocolVersion = BitConverter.ToInt32(buffer, 0);
            if (ProtocolVersion != protocolVersion)
                throw new AgentConnectionException("Invalid handshake received: unsupported protocol version.");

            return BitConverter.ToInt32(buffer, 4);
        }
        public async Task SendHandshakeResponse(int sessionId)
        {
            this.Stream.WriteByte((byte)AgentConfig.Encryption);
            await this.Stream.FlushAsync();

            var mode = AgentConfig.Encryption;
            if (mode == AgentEncryptionMode.Ssl)
            {
                var certificateName = AgentConfig.CertificateName;

                var store = new X509Store(StoreLocation.LocalMachine);
                store.Open(OpenFlags.OpenExistingOnly | OpenFlags.ReadOnly);
                var cert = store
                    .Certificates
                    .Cast<X509Certificate2>()
                    .Where(c => c.FriendlyName == certificateName)
                    .FirstOrDefault();

                var stream = new SslStream(this.Stream, false);
                stream.AuthenticateAsServer(cert);
                this.Stream = stream;
            }
            else if (mode == AgentEncryptionMode.Aes)
            {
                var initialIV = new byte[16];
                using (var random = new RNGCryptoServiceProvider())
                {
                    random.GetBytes(initialIV);
                }

                this.messageFormatter = new MessageFormatter.AesMessageFormatter(AgentConfig.EncryptionKey, initialIV);
                await this.Stream.WriteAsync(initialIV, 0, 16);
            }

            var buffer = new MemoryStream(100);
            var writer = new BinaryWriter(buffer);
            writer.Write(sessionId);

            var version = typeof(InedoAgentServer).Assembly.GetName().Version;
            writer.Write(version.Major);
            writer.Write(version.Minor);
            writer.Write(version.Build);
            writer.Write(version.Revision);

            buffer.Position = 0;
            await buffer.CopyToAsync(this.Stream);

            this.IsConnected = true;
        }
        public async Task<AgentMessage> ReadMessageAsync()
        {
            if (!this.IsConnected)
                throw new InvalidOperationException("Channel is not connected.");

            var lengthBuffer = new byte[sizeof(int)];
            int read;
            try
            {
                read = await this.Stream.ReadBlockAsync(lengthBuffer, 0, sizeof(int));
            }
            catch (ObjectDisposedException)
            {
                return null;
            }

            if (read == 0)
            {
                this.OnDisconnected(true);
                return null;
            }

            if (read < sizeof(int))
                throw new AgentConnectionException();

            int length = BitConverter.ToInt32(lengthBuffer, 0);
            if (length < 0)
                throw new AgentConnectionException();

            var buffer = new byte[length];
            read = await this.Stream.ReadBlockAsync(buffer, 0, buffer.Length);
            if (read != buffer.Length)
                throw new AgentConnectionException();

            return this.messageFormatter.GetIncomingMessage(new MemoryStream(buffer, false));
        }
        public async Task SendMessageAsync(int messageTypeCode, Stream messageData)
        {
            using (var outgoingStream = this.messageFormatter.GetOutgoingMessageStream(messageTypeCode, messageData))
            {
                await this.Stream.WriteAsync(BitConverter.GetBytes((int)outgoingStream.Length), 0, sizeof(int));
                await outgoingStream.CopyToAsync(this.Stream);
                await this.Stream.FlushAsync();
            }
        }

        private abstract class MessageFormatter : IDisposable
        {
            protected MessageFormatter()
            {
            }

            public abstract Stream GetOutgoingMessageStream(int messageTypeCode, Stream messageData);
            public abstract AgentMessage GetIncomingMessage(Stream messageData);
            public virtual void Dispose()
            {
            }

            public sealed class StandardMessageFormatter : MessageFormatter
            {
                public override Stream GetOutgoingMessageStream(int messageTypeCode, Stream messageData)
                {
                    var stream = new MemoryStream((int)(messageData.Length - messageData.Position) + sizeof(int));
                    stream.Write(BitConverter.GetBytes(messageTypeCode), 0, sizeof(int));
                    messageData.CopyTo(stream);
                    stream.Position = 0;
                    return stream;
                }
                public override AgentMessage GetIncomingMessage(Stream messageData)
                {
                    var reader = new BinaryReader(messageData);
                    int commandCode = reader.ReadInt32();
                    var agentType = (AgentType)reader.ReadByte();
                    return new AgentMessage(agentType, commandCode, new OffsetStream(messageData, messageData.Position));
                }
            }

            public sealed class AesMessageFormatter : MessageFormatter
            {
                private Aes decryption;
                private Aes encryption;
                private RNGCryptoServiceProvider random = new RNGCryptoServiceProvider();

                public AesMessageFormatter(byte[] key, byte[] iv)
                {
                    this.decryption = GetAesProvider();
                    this.decryption.Key = key;
                    this.decryption.IV = iv;

                    this.encryption = GetAesProvider();
                    this.encryption.Key = key;
                    this.encryption.IV = iv;
                }

                public override Stream GetOutgoingMessageStream(int messageTypeCode, Stream messageData)
                {
                    lock (this.encryption)
                    {
                        var nextKey = new byte[16];
                        this.random.GetBytes(nextKey);

                        try
                        {
                            var stream = new MemoryStream();

                            using (var encryptor = this.encryption.CreateEncryptor())
                            using (var cryptoStream = new CryptoStream(new UndisposableStream(stream), encryptor, CryptoStreamMode.Write))
                            {
                                cryptoStream.Write(nextKey, 0, 16);
                                new BinaryWriter(cryptoStream).Write(messageTypeCode);
                                messageData.CopyTo(cryptoStream);
                            }

                            stream.Position = 0;
                            return stream;
                        }
                        finally
                        {
                            this.encryption.IV = nextKey;
                        }
                    }
                }
                public override AgentMessage GetIncomingMessage(Stream messageData)
                {
                    lock (this.decryption)
                    {
                        var nextKey = new byte[16];
                        try
                        {
                            using (var decryptor = this.decryption.CreateDecryptor())
                            using (var cryptoStream = new CryptoStream(messageData, decryptor, CryptoStreamMode.Read))
                            {
                                cryptoStream.Read(nextKey, 0, 16);
                                var buffer = new MemoryStream();
                                cryptoStream.CopyTo(buffer);
                                buffer.Position = 0;

                                int commandCode = new BinaryReader(buffer).ReadInt32();
                                var agentType = (AgentType)buffer.ReadByte();

                                return new AgentMessage(agentType, commandCode, new OffsetStream(buffer, buffer.Position));
                            }
                        }
                        finally
                        {
                            this.decryption.IV = nextKey;
                        }
                    }
                }
                public override void Dispose()
                {
                    this.decryption.Dispose();
                    this.encryption.Dispose();
                    this.random.Dispose();
                }
            }
        }
    }
}
